package com.redhat.thermostat.tools.eclipse.plugin.wizards;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.commands.NotEnabledException;
import org.eclipse.core.commands.NotHandledException;
import org.eclipse.core.commands.common.NotDefinedException;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IWorkspaceRoot;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Status;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IPackageFragment;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.handlers.IHandlerService;

import com.redhat.thermostat.tools.eclipse.plugin.Activator;
import com.redhat.thermostat.tools.eclipse.plugin.wizards.PluginXmlCreator.Decisions;

/** Creates a Thermostat project using the preferred thermostat defaults, settings, and style */
public class ProjectCreator {

    private static final String POM_FILE_NAME = "pom.xml";
    private static final String POM_HEADER = ""
            + "<?xml version='1.0' encoding='UTF-8'?>\n"
            + "<project xsi:schemaLocation='http://maven.apache.org/POM/4.0.0\n"
            + "                             http://maven.apache.org/xsd/maven-4.0.0.xsd'\n"
            + "         xmlns='http://maven.apache.org/POM/4.0.0'\n"
            + "         xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance'>\n"
            + "\n"
            + "  <modelVersion>4.0.0</modelVersion>\n"
            + "\n";
    private static final String POM_FOOTER = ""
            + "</project>\n";

    private static final String OSGI_DEPENDENCIES = ""
            + "    <dependency>\n"
            + "      <groupId>org.osgi</groupId>\n"
            + "      <artifactId>org.osgi.core</artifactId>\n"
            + "    </dependency>\n"
            + "    <dependency>\n"
            + "      <groupId>org.osgi</groupId>\n"
            + "      <artifactId>org.osgi.compendium</artifactId>\n"
            + "    </dependency>\n";

    private static final String BUNDLE_ACTIVATOR_NAME = "Activator";
    private static final String BUNDLE_ACTIVATOR = ""
            + "package ${package.name};\n" // ${package.name} is replaced with actual package name
            + "\n"
            + "import org.osgi.framework.BundleActivator;\n"
            + "import org.osgi.framework.BundleContext;\n"
            + "\n"
            + "public class " + BUNDLE_ACTIVATOR_NAME + " implements BundleActivator {\n"
            + "\n"
            + "    public void start(BundleContext context) throws Exception {\n"
            + "        // TODO Auto-generated method stub\n"
            + "    }\n"
            + "\n"
            + "    public void stop(BundleContext context) throws Exception {\n"
            + "        // TODO Auto-generated method stub\n"
            + "    }\n"
            + "\n"
            + "}\n";

    private IHandlerService serviceHandler;

    private String packagePrefix;
    private String groupId;
    private String artifactId;
    private String version;
    private String thermostatVersion;

    private String parentId;

    public ProjectCreator(String groupId, String artifactId, String version, String packagePrefix, String thermostatVersion) {
        this((IHandlerService) PlatformUI.getWorkbench().getService(IHandlerService.class), groupId, artifactId, version, packagePrefix, thermostatVersion);
    }

    public ProjectCreator(IHandlerService handler, String groupId, String artifactId, String version, String packagePrefix, String thermostatVersion) {
        this.serviceHandler = handler;
        this.groupId = groupId;
        this.artifactId = artifactId;
        this.version = version;
        this.packagePrefix = packagePrefix;

        this.thermostatVersion = thermostatVersion;

        this.parentId = artifactId + "-parent";

    }

    public void create() throws CoreException, IOException {
        createParentModuleProject();

        createCommonSubproject();

        createJavaSubproject("agent");
        createJavaSubproject("client");
        createJavaSubproject("cli");

        createDistributionProject();

        refreshProjects();
    }

    private void createParentModuleProject() throws CoreException, IOException {
        Projects projects = createJavaProject(parentId);

        String pomContents = ""
                + POM_HEADER
                + "  <groupId>" + groupId + "</groupId>\n"
                + "  <artifactId>" + parentId + "</artifactId>\n"
                + "  <version>" + version + "</version>\n"
                + "  <packaging>pom</packaging>\n"
                + "\n"
                + "  <modules>\n"
                + "    <module>" + this.artifactId + "-agent</module>\n"
                + "    <module>" + this.artifactId + "-common</module>\n"
                + "    <module>" + this.artifactId + "-client</module>\n"
                + "    <module>" + this.artifactId + "-cli</module>\n"
                + "    <module>" + this.artifactId + "-distribution</module>\n"
                + "  </modules>\n"
                + "\n"
                + "  <dependencyManagement>\n"
                + "    <dependencies>\n"
                + "      <dependency>\n"
                + "        <groupId>org.osgi</groupId>\n"
                + "        <artifactId>org.osgi.core</artifactId>\n"
                + "        <version>4.3.1</version>\n"
                + "      </dependency>\n"
                + "      <dependency>\n"
                + "        <groupId>org.osgi</groupId>\n"
                + "        <artifactId>org.osgi.compendium</artifactId>\n"
                + "        <version>4.3.1</version>\n"
                + "      </dependency>\n"
                + "    </dependencies>\n"
                + "  </dependencyManagement>\n"
                + POM_FOOTER;

        createPomFile(projects, pomContents);
    }

    private void createCommonSubproject() throws IOException, CoreException {
        String suffix = "common";
        String artifactId = this.artifactId + "-" + suffix;
        String packageName = packagePrefix + "." + suffix;
        Projects projects = createJavaCodeProject(artifactId, packageName);

        String pomContents = ""
                + POM_HEADER
                + "\n"
                + "  <parent>\n"
                + "    <groupId>" + groupId + "</groupId>\n"
                + "    <artifactId>" + parentId + "</artifactId>\n"
                + "    <version>" + version + "</version>\n"
                + "  </parent>\n"
                + "\n"
                + "  <artifactId>" + artifactId + "</artifactId>\n"
                + "  <packaging>jar</packaging>\n"
                + "\n"
                + "  <dependencies>\n"
                + OSGI_DEPENDENCIES
                + "  </dependencies>\n"
                + POM_FOOTER;

        createPomFile(projects, pomContents);

    }

    private void createJavaSubproject(String subproject) throws CoreException, IOException {
        String artifactId = this.artifactId + "-" + subproject;
        String subpackage = this.packagePrefix + "." + subproject;

        Projects projects = createJavaCodeProject(artifactId, subpackage);

        String pomContents = ""
                + POM_HEADER
                + "  <parent>\n"
                + "    <groupId>" + groupId + "</groupId>\n"
                + "    <artifactId>" + parentId + "</artifactId>\n"
                + "    <version>" + version + "</version>\n"
                + "  </parent>\n"
                + "\n"
                + "  <artifactId>" + artifactId + "</artifactId>\n"
                + "  <packaging>jar</packaging>\n"
                + "\n"
                + "  <dependencies>\n"
                + "    <dependency>\n"
                + "      <groupId>" + groupId + "</groupId>\n"
                + "      <artifactId>" + this.artifactId + "-common</artifactId>\n"
                + "      <version>" + version + "</version>\n"
                + "    </dependency>\n"
                + OSGI_DEPENDENCIES
                + "  </dependencies>\n"
                + POM_FOOTER;

        createPomFile(projects, pomContents);
    }

    private Projects createJavaCodeProject(String projectName, String packageName) throws CoreException {
        Projects projects = createJavaProject(projectName);

        IProject project = projects.project;

        IFolder srcFolder = project.getFolder("src");
        srcFolder.create(false, true, null);

        IFolder srcMainFolder = srcFolder.getFolder("main");
        srcMainFolder.create(true, true, null);

        IFolder srcMainJavaFolder = srcMainFolder.getFolder("java");
        srcMainJavaFolder.create(true, true, null);

        IFolder srcTestFolder = srcFolder.getFolder("test");
        srcTestFolder.create(true, true, null);

        IFolder srcTestJavaFolder = srcTestFolder.getFolder("java");
        srcTestJavaFolder.create(true, true, null);

        createBundleActivator(projects, srcMainJavaFolder, packageName);

        return projects;
    }

    private void createBundleActivator(Projects projects, IFolder root, String packageName) throws CoreException {
        IJavaProject javaProject = projects.javaProject;

        IPackageFragment pack = javaProject.getPackageFragmentRoot(root)
                                    .createPackageFragment(packageName, false, null);

        String bundleActivatorContents = BUNDLE_ACTIVATOR.replace("${package.name}", pack.getElementName());
        pack.createCompilationUnit(BUNDLE_ACTIVATOR_NAME + ".java", bundleActivatorContents, false, null);
    }

    private void createDistributionProject() throws CoreException, IOException {
        String artifactId = this.artifactId + "-distribution";

        Projects projects = createJavaProject(artifactId);

        String pomContents = ""
                + POM_HEADER
                + "  <parent>\n"
                + "    <groupId>" + groupId + "</groupId>\n"
                + "    <artifactId>" + parentId + "</artifactId>\n"
                + "    <version>" + version + "</version>\n"
                + "  </parent>\n"
                + "\n"
                + "  <artifactId>" + artifactId + "</artifactId>\n"
                + "  <packaging>jar</packaging>\n"
                + "\n"
                + "  <!-- Explicitly list all plug-in artifacts, transitive dependencies\n"
                + "       are not included in assembly. -->\n"
                + "  <dependencies>\n"
                + "    <dependency>\n"
                + "      <groupId>" + groupId + "</groupId>\n"
                + "      <artifactId>" + this.artifactId + "-agent</artifactId>\n"
                + "      <version>${project.version}</version>\n"
                + "    </dependency>\n"
                + "    <dependency>\n"
                + "      <groupId>" + groupId + "</groupId>\n"
                + "      <artifactId>" + this.artifactId + "-common</artifactId>\n"
                + "      <version>${project.version}</version>\n"
                + "    </dependency>\n"
                + "    <dependency>\n"
                + "      <groupId>" + groupId + "</groupId>\n"
                + "      <artifactId>" + this.artifactId + "-client</artifactId>\n"
                + "      <version>${project.version}</version>\n"
                + "    </dependency>\n"
                + "    <dependency>\n"
                + "      <groupId>" + groupId + "</groupId>\n"
                + "      <artifactId>" + this.artifactId + "-cli</artifactId>\n"
                + "      <version>${project.version}</version>\n"
                + "    </dependency>\n"
                + "  </dependencies>\n"
                + POM_FOOTER;

        createPomFile(projects, pomContents);

        IProject project = projects.project;

        IFolder srcFolder = project.getFolder("src");
        srcFolder.create(false, true, null);

        IFolder srcMainFolder = srcFolder.getFolder("main");
        srcMainFolder.create(true, true, null);

        IFolder srcMainResourcesFolder = srcMainFolder.getFolder("resources");
        srcMainResourcesFolder.create(true, true, null);

        // FIXME enable this
        // createAssemblyDescriptor();
        createThermostatPluginXml(project.getName());

    }

    private void createThermostatPluginXml(String projectName) throws CoreException, IOException {
        PluginXmlCreator creator = new PluginXmlCreator(ResourcesPlugin.getWorkspace().getRoot());
        creator.create(projectName, new NullProgressMonitor(), new Decisions() {
            @Override
            public boolean overwriteExistingFile() {
                return true;
            }
        });
    }

    private Projects createJavaProject(String projectName) throws CoreException {
        IWorkspaceRoot root = ResourcesPlugin.getWorkspace().getRoot();
        IProject project = root.getProject(projectName);
        if (project.exists()) {
            project.delete(true, null);
        }
        try {
            project.create(null);
            project.open(null);
        } catch (CoreException e) {
            e.printStackTrace();
        }

        addJavaAndMavenNature(project);

        // Create Java project
        IJavaProject javaProject = JavaCore.create(project);

        return new Projects(project, javaProject);
    }

    private void addJavaAndMavenNature(IProject project) throws CoreException {
        IProjectDescription description = project.getDescription();
        addJavaAndMavenNature(description);
        project.setDescription(description, null);
    }

    private void addJavaAndMavenNature(IProjectDescription description) {
        String[] existingNatureIds = description.getNatureIds();
        Set<String> natures = new TreeSet<>(Arrays.asList(existingNatureIds));
        natures.add(JavaCore.NATURE_ID);
        // provided by m2e
        // TODO what's the symbolic constant for this? IMavenConstants.NATURE_ID?
        natures.add("org.eclipse.m2e.core.maven2Nature");
        String[] newNatureIds = natures.toArray(new String[natures.size()]);
        description.setNatureIds(newNatureIds);
    }

    private void createPomFile(Projects projects, String contents) throws IOException, CoreException {
        IProject project = projects.project;
        IFile pomFile = project.getFile(POM_FILE_NAME);
        try (InputStream contentStream = new ByteArrayInputStream(contents.getBytes(StandardCharsets.UTF_8))) {
            if (pomFile.exists()) {
                pomFile.setContents(contentStream, true, true, null);
            } else {
                pomFile.create(contentStream, true, null);
            }
        }
    }

    private void refreshProjects() throws CoreException {
        // Refresh projects
        ResourcesPlugin.getWorkspace().getRoot().refreshLocal(IWorkspaceRoot.DEPTH_INFINITE, new NullProgressMonitor());

        // Refresh again, just to make sure changes are sane by the time maven sees them
        ResourcesPlugin.getWorkspace().getRoot().refreshLocal(IWorkspaceRoot.DEPTH_INFINITE, new NullProgressMonitor());

        // Refresh maven things
        try {
            // provided by m2e
            final String UPADATE_MAVEN_PROJECT_COMMAND = "org.eclipse.m2e.core.ui.command.updateProject";
            serviceHandler.executeCommand(UPADATE_MAVEN_PROJECT_COMMAND, null);
        } catch (ExecutionException | NotDefinedException | NotEnabledException | NotHandledException e) {
            Status s = new Status(IStatus.ERROR, Activator.PLUGIN_ID, "Error refreshing maven configuration", e);
            throw new CoreException(s);
        }
    }

    private static class Projects {
        private final IProject project;
        private final IJavaProject javaProject;

        public Projects(IProject project, IJavaProject javaProject) {
            this.project = project;
            this.javaProject = javaProject;
        }
    }
}
